/*
  Lab Solutions for Lab 9 Task 2
  
  Created by Michael Bykov (michael.bykov@wsu.edu)
  
  The interface for the binary search tree class.
  Much of the implementation for our binary search tree is recursive.
*/

#pragma once
#include "BSTNode.h"

// A templated binary search tree class so that we can have a tree with values of any type.
template<typename T>
class BST {
public:
  
  // Default constructor.
  BST();
  
  // Copy constructor. (rule of three)
  BST(const BST<T> &other);
  
  // Assignment operator. (rule of three)
  BST &operator =(const BST<T> &other);
  
  // Destructor. (rule of three)
  ~BST();
  
  
  
  // Determine if the tree is empty.
  //
  // Time Complexity:
  // Since we aren't touching any of the elements in the tree, the time complexity is unrelated
  // to the number of elements in the tree therefore it is constant: O(1).
  bool isEmpty() const;
  
  
  
  // Insert a new value into the tree.
  //
  // Time Complexity:
  // Depending on how our tree is laid out, our time complexity can vary significantly.
  // For a balanced tree (the height of the left side of the tree is the same as the height of the right side of the tree),
  // the time complexity can be as low as O(log n) because every time we recursively insert in a subtree, we are cutting the
  // number of nodes we have to look at in half (since in a balanced tree each side of every node should have approximately the same number of nodes).
  // For a completely unbalanced tree (think linked list where we have a root node and all of the nodes are inserted to the right) the time complexty
  // can be as high as O(n) because we may have to access every single node before we find an appropriate place to insert our new node.
  void insert(const T &value);
  
  
  
  // Print the tree in order.
  //
  // Time Complexity:
  // Since we have to visit every node in the tree, the time complexity is O(n) where n is the number of nodes in the tree.
  void inOrderTraversal();
  
  // Print the tree in pre-order.
  //
  // Time Complexity:
  // Same as above.
  void preOrderTraversal();
  
  // Print the tree in post-order.
  //
  // Time Complexity:
  // Same as above.
  void postOrderTraversal();
  
  
  
  // For task 3:
  // Store the contents of the tree in a vector in sorted order by traversing the tree in order.
  //
  // Time Complexity:
  // Since this is just performing an in-order traversal, the time complexity is O(n) since we are accessing all of the nodes and there are n nodes in the tree.
  void sort(std::vector<T> &values);
  
  
  
  // Access the root node of the tree.
  BSTNode<T> *root();
  
  // Change the root node of the tree.
  void setRoot(BSTNode<T> *root);
  
  
  
private:
  
  // The root node of the tree.
  BSTNode<T> *_root;
  
  
  
  // Recursively delete the tree.
  //
  // Time Complexity:
  // Since we have to visit every node in the tree, the time complexity is O(n) where n is the number of nodes in the tree.
  void _deleteTree(BSTNode<T> *node);
  
  // Recursively copy the tree.
  //
  // Time Complexity:
  // Same as above.
  BSTNode<T> *_copyTree(BSTNode<T> *node);
};



// Print a binary search tree.
template<typename T>
std::ostream &operator<<(std::ostream &stream, const BST<T> &bst);



// Include the implementation file like what we did in BSTNode.h.
#include "BST.tpp"
