/*
  Lab Solutions for Lab 9 Task 2
  
  Created by Michael Bykov (michael.bykov@wsu.edu)
  
  The implementation of the binary search tree class.
  Much of the implementation for our binary search tree is recursive.
*/

#pragma once
#include "BST.h"

template<typename T>
BST<T>::BST() {
  // Start with an empty tree
  _root = nullptr;
}

template<typename T>
BST<T>::BST(const BST<T> &other) {
  // Copy the other tree recursively
  if (other._root != nullptr) {
    _root = _copyTree(_root, other._root);
  } else {
    _root = nullptr;
  }
}

template<typename T>
BST<T> &BST<T>::operator=(const BST<T> &other) {
  // Delete the tree we currently have
  deleteTree(_root);
  
  // Copy the other tree recursively
  if (other._root != nullptr) {
    _root = _copyTree(_root, other._root);
  } else {
    _root = nullptr;
  }
  return *this;
}

template<typename T>
BST<T>::~BST() {
  // Delete the tree recursively
  _deleteTree(_root);
  
  // Make sure we don't have any dangling pointers
  _root = nullptr;
}



template<typename T>
bool BST<T>::isEmpty() const {
  // The tree is empty when we don't have any nodes
  return _root == nullptr;
}



template<typename T>
void BST<T>::insert(const T &value) {
  // Create the new node
  BSTNode<T> *newNode = new BSTNode<T>(value);
  
  // Determine how to insert the node
  if (_root == nullptr) {
    // First node in the tree: set the root similar to what we did when inserting into an empty linked list
    _root = newNode;
  } else {
    // Insert recursively
    _root->insertNode(newNode);
  }
}



template<typename T>
void BST<T>::inOrderTraversal() {
  // Recursive in-order traversal
  if (_root != nullptr) {
    _root->inOrderTraversal();
  }
}

template<typename T>
void BST<T>::preOrderTraversal() {
  // Recursive pre-order traversal
  if (_root != nullptr) {
    _root->preOrderTraversal();
  }
}

template<typename T>
void BST<T>::postOrderTraversal() {
  // Recursive post-order traversal
  if (_root != nullptr) {
    _root->postOrderTraversal();
  }
}



template<typename T>
void BST<T>::sort(std::vector<T> &values) {
  // Recursive in-order traversal to access the tree in sorted order
  if (_root != nullptr) {
    _root->sort(values);
  }
}



template<typename T>
BSTNode<T> *BST<T>::root() {
  return _root;
}

template<typename T>
void BST<T>::setRoot(BSTNode<T> *root) {
  _root = root;
}



template<typename T>
void BST<T>::_deleteTree(BSTNode<T> *node) {
  // Check that we have something to delete
  if (node != nullptr) {
    // Delete the children recursively
    _deleteTree(node->left());
    _deleteTree(node->right());
    
    // Delete the current node
    delete node;
  }
}

template<typename T>
BSTNode<T> *BST<T>::_copyTree(BSTNode<T> *node) {
  // Create a copy of the other node
  BSTNode<T> *copy = new BSTNode<T>(node->value());
  
  // Copy the children recursively
  if (node->left() != nullptr) {
    copy->setLeft(_copyTree(node->left()));
  }
  if (node->right() != nullptr) {
    copy->setRight(_copyTree(node->right()));
  }
  
  return copy;
}
