/*
  Lab Solutions for Lab 9 Task 1
  
  Created by Michael Bykov (michael.bykov@wsu.edu)
  
  The interface of the BSTNode class.
  Just like we do with other classes, we separate the interface from the implementation by putting the implementation in a separate file.
  With templates, since the implementation has to be in the header file, we include our source file (in this case "BSTNode.tpp") at the end of our class interface
  so that the implementation can be available with the interface.
  Since we are including the source file in this header, we can choose a different file extension for our source file so that it won't be treated as a source
  file directly, but rather a header file.
  Some common extension for implementation files for templates are ".tpp", "ipp", and "inl" but you can also use just a ".h" extension.
*/

#pragma once
#include <iostream>
#include <vector>

/*
  A templated BST node class.
  Whenever we create an instance of this class in our code, C++ will substitute in the template argument we provide for
  any T that it sees in our interface or implementation.
*/
template<typename T>
class BSTNode {
public:
  
  // A constructor that stores a value in the node.
  BSTNode(const T &value);
  
  
  
  // Access the value stored in the node.
  const T &value() const;
  
  // Change the value stored in the node.
  void setValue(const T &value);
  
  // Access the left child of the node.
  BSTNode<T> *left();
  
  // Change the left child of the node.
  void setLeft(BSTNode<T> *left);
  
  // Access the right child of the node.
  BSTNode<T> *right();
  
  // Change the right child of the node.
  void setRight(BSTNode<T> *right);
  
  
  
  // Insert a new node into the tree recursively.
  void insertNode(BSTNode<T> *newNode);
  
  
  
  // Print the tree in order.
  void inOrderTraversal();
  
  // Print the tree in pre-order.
  void preOrderTraversal();
  
  // Print the tree in post-order.
  void postOrderTraversal();
  
  
  
  // For task 3:
  // Traverse the tree in order and store the values in a vector.
  // The vector will contain the contents of the tree in sorted order by the properties of a binary search tree.
  void sort(std::vector<T> &values);
  
  
  
private:
  
  // The value stored in the node.
  T _value;
  
  // The left child of the node.
  BSTNode<T> *_left;
  
  // The right child of the node.
  BSTNode<T> *_right;
};



// Print a node to a stream.
template<typename T>
std::ostream &operator<<(std::ostream &stream, BSTNode<T> &node);



// Include the implementation of the template class so that it is available to the compiler when it needs to substitute in the template argument for T.
#include "BSTNode.tpp"
