/*
  Lab Solutions for Lab 9 Task 1
  
  Created by Michael Bykov (michael.bykov@wsu.edu)
  
  The implementation of the BSTNode class.
  This file is #include'd in the BSTNode.h file so that any values of T that are used in our implementation
  can be changed to the appropriate template argument when a new instance of the class is used in our code.
*/

#pragma once
#include "BSTNode.h"

template<typename T>
BSTNode<T>::BSTNode(const T &value) {
  // Store the value in the node
  _value = value;
  // We don't have any children yet
  _left = nullptr;
  _right = nullptr;
}



template<typename T>
const T &BSTNode<T>::value() const {
  // Access the value stored in the node
  return _value;
}

template<typename T>
void BSTNode<T>::setValue(const T &value) {
  // Change the value stored in the node
  _value = value;
}

template<typename T>
BSTNode<T> *BSTNode<T>::left() {
  return _left;
}

template<typename T>
void BSTNode<T>::setLeft(BSTNode<T> *left) {
  _left = left;
}

template<typename T>
BSTNode<T> *BSTNode<T>::right() {
  return _right;
}

template<typename T>
void BSTNode<T>::setRight(BSTNode<T> *right) {
  _right = right;
}



template<typename T>
void BSTNode<T>::insertNode(BSTNode<T> *newNode) {
  // Determine where to insert the new node in relation to this node
  // Remember that a binary search tree is set up such that all the values to the left of any node are less than the node itself
  // and all the values to the right of any node are greater than the node itself
  // Which means that all of the values in the left subtree of any node are less than the values in the right subtree of that node
  // So, for instance, if we have a node whose value is less than tht of the current node we should insert it in the left subtree
  // so that the integrity of the binary search tree is maintained
  if (newNode->value() < _value) {
    // Value is less than the current node: insert in the left subtree
    if (_left == nullptr) {
      // No left subtree so we can insert the node directly here
      _left = newNode;
    } else {
      // There is a left subtree so we need to insert the node into that subtree
      _left->insertNode(newNode);
    }
  } else {
    // Values is more than the current node: insert in the right subtree
    if (_right == nullptr) {
      // No right subtree so we can insert the node directly here
      _right = newNode;
    } else {
      // There is a right subtree so we need to insert the node into that subtree
      _right->insertNode(newNode);
    }
  }
}



/*
  When traversing a tree, there are three main orders for visiting the nodes:
  - In-order: we visit the left subtree, then the node itself, then the right subtree
  - Pre-order: we visit the node itself, then the left subtree, then the right subtree
  - Post-order: we visit the left subtree, then the right subtree, then the node itself
  
  An easy way of remembering and understanding the orderings is to think of them in relation to the node itself.
  
  In [pre]-order the node goes before the subtree (pre = node before subtrees)
  i.e. node, left subtree, right subtree  
  
  In [post]-order the node goes after the subtree (post = node after subtrees)
  i.e. left subtree, right subtree, node
  
  In [in]-order the node goes in between the subtrees (in(-between) = node in between subtrees)
  i.e. left subtree, node, right subtree
*/

template<typename T>
void BSTNode<T>::inOrderTraversal() {
  // In-order: left, node, right
  
  // Traverse the left subtree
  if (_left != nullptr) {
    _left->inOrderTraversal();
  }
  
  // Print the value of the current node
  std::cout << _value << " ";
  
  // Traverse the right subtree
  if (_right != nullptr) {
    _right->inOrderTraversal();
  }
}

template<typename T>
void BSTNode<T>::preOrderTraversal() {
  // Pre-order: node, left, right
  
  // Print the value of the current node
  std::cout << _value << " ";
  
  // Traverse the left subtree
  if (_left != nullptr) {
    _left->preOrderTraversal();
  }
  
  // Traverse the right subtree
  if (_right != nullptr) {
    _right->preOrderTraversal();
  }
}

template<typename T>
void BSTNode<T>::postOrderTraversal() {
  // Post-order: left, right, node
  
  // Traverse the left subtree
  if (_left != nullptr) {
    _left->postOrderTraversal();
  }
  
  // Traverse the right subtree
  if (_right != nullptr) {
    _right->postOrderTraversal();
  }
  
  // Print the value of the current node
  std::cout << _value << " ";
}



template<typename T>
void BSTNode<T>::sort(std::vector<T> &values) {
  // Given that we know that all of the values on the left side of any node are less than the value of that node itself
  // and all of the values on the right side of any node are greater than the value of that node itself,
  // if we traverse the tree in order (that is, we visit the left subtree, then the node itself, then the right subtree)
  // then we will visit all of the nodes in the tree in there sorted order so if we store the values in a vector as we visit them
  // then the vector will contain the values of the tree in sorted order
  
  // Traverse the left subtree
  if (_left != nullptr) {
    _left->sort(values);
  }
  
  // Store the value of the current node
  values.push_back(_value);
  
  // Traverse the right subtree
  if (_right != nullptr) {
    _right->sort(values);
  }
}



template<typename T>
std::ostream &operator<<(std::ostream &stream, BSTNode<T> &node) {
  // An in-order traversal of the tree is a natural way to print the tree
  
  // Print the left subtree
  if (node.left() != nullptr) {
    stream << *node.left() << ", ";
  }
  
  // Print the value of the current node
  stream << node.value();
  
  // Print the right subtree
  if (node.right() != nullptr) {
    stream << ", " << *node.right();
  }
  
  // Always remember to return the stream at the end
  return stream;
}
